/* ***************************************************************************
 * $Id$
 *
 * Purpose      : GDAL Image Splitting Program
 * Author       : Homme Zwaagstra, geodata@soton.ac.uk
 * Maintenance  : Mateusz Loskot, mateusz@loskot.net
 *
 * This program is heavily derived from the GDAL gdal_translate utility
 *
 * ***************************************************************************
 * Copyright (c) 2007, Mateusz Loskot
 * Copyright (c) 2004, Homme Zwaagstra
 * Copyright (c) 1998, 2002, Frank Warmerdam
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include "gdal_priv.h"
#include <cpl_conv.h>
#include <cpl_string.h>
#include <ogr_spatialref.h>
#include <vrt/vrtdataset.h>

/**********************************************************************/
/*            Anonymous namespace with private utilities              */
/**********************************************************************/
namespace
{

const int DefaultColsNumber = 2;
const int DefaultRowsNumber = 2;

#ifdef WIN32
const char PathSeparator = '\\';
#else
const char PathSeparator = '/';
#endif

template <typename T>
inline int sgn(T value)
{
    return ( ( value ) < 0 ? -1 : ( ( value ) > 0 ? 1 : 0 ) );
}

template <typename R, typename T>
inline R round(T value)
{
    return static_cast<R>( ( value ) + ( 0.5 * sgn( value ) ) );
}

/**********************************************************************/
/*                           Usage()                                  */
/**********************************************************************/

void Usage()
{
    int	iDr;

    printf( "Usage: gdalsplit [--version]\n"
        "       [-c cols]\n"
        "       [-r rows]\n"
        "       [-quiet]\n"
        "       src_dataset...\n\n" );

    printf( "%s\n\n", GDALVersionInfo( "--version" ) );
    printf( "The following format drivers are configured and support output:\n" );
    for( iDr = 0; iDr < GDALGetDriverCount(); iDr++ )
    {
        GDALDriverH hDriver = GDALGetDriver(iDr);

        if( GDALGetMetadataItem( hDriver, GDAL_DCAP_CREATE, NULL ) != NULL
            || GDALGetMetadataItem( hDriver, GDAL_DCAP_CREATECOPY,
            NULL ) != NULL )
        {
            printf( "  %s: %s\n",
                GDALGetDriverShortName( hDriver ),
                GDALGetDriverLongName( hDriver ) );
        }
    }
}

/**********************************************************************/
/*                           CreateString()                           */
/**********************************************************************/

char* CreateString(const char* fmt, ...)
{
    /* Guess we need no more than 100 bytes. */
    int     n = 0;
    size_t  size = 100;
    char*   p = NULL;
    va_list ap;

    if( NULL == (p = (char *)CPLMalloc( size ) ) )
        return NULL;

    while (1)
    {
        /* Try to print in the allocated space. */
        va_start( ap, fmt );
        n = vsnprintf( p, size, fmt, ap );
        va_end( ap );

        /* If that worked, return the string. */
        if( n > -1 && n < size )
            return p;

        /* Else try again with more space. */
        if( n > -1 )    /* glibc 2.1 */
            size = n+1; /* precisely what is needed */
        else           /* glibc 2.0 */
            size *= 2;  /* twice the old size */

        if( NULL == (p = (char*)CPLRealloc( p, size ) ) )
            return NULL;
    }
}

/**********************************************************************/
/*                           CreateFileName()                         */
/**********************************************************************/

char* CreateFileName( const char* pszSource, int col, int row )
{
    char* base = CPLStrdup( CPLGetBasename( pszSource ) );
    char* ext = CPLStrdup( CPLGetExtension( pszSource ) );
    char* dir = CPLStrdup( CPLGetPath( pszSource ) );
    CPLAssert( NULL != base && NULL != ext && NULL != dir );

    char* pszFilename = NULL;

    const char* path = dir;
    if( strlen( dir ) <= 0 )
        path = ".";

    if( NULL != ext )
    {
        pszFilename = CreateString( "%s%c%s_c%d_r%d.%s", path,
                                    PathSeparator, base,
                                    col, row, ext );
    }
    else
    {
        pszFilename = CreateString( "%s%c%s_c%d_r%d", path,
                                    PathSeparator, base, col, row );
    }

    CPLFree( dir );
    CPLFree( ext );
    CPLFree( base );

    return pszFilename;
}

/**********************************************************************/
/*                           AttachMetadata()                         */
/**********************************************************************/

void AttachMetadata( GDALDatasetH hDS, char **papszMetadataOptions )
{
    int nCount = CSLCount(papszMetadataOptions);

    for( int i = 0; i < nCount; i++ )
    {
        char* pszKey = NULL;
        const char* pszValue = NULL;

        pszValue = CPLParseNameValue( papszMetadataOptions[i], &pszKey );
        GDALSetMetadataItem( hDS, pszKey, pszValue, NULL );

        CPLFree( pszKey );
    }

    CSLDestroy( papszMetadataOptions );
}					

/**********************************************************************/
/*                           CopyRasterBand()                         */
/**********************************************************************/

CPLErr CopyRasterBand( GDALDataset* poSrcDS, GDALDataset *poDstDS,
                             int nBand,
                             int nSrcXOff, int nSrcYOff,
                             int nSrcXSize, int nSrcYSize )
{
    VRTSourcedRasterBand*   poVRTBand = NULL;
    GDALRasterBand*         poSrcBand = NULL;
    GDALDataType eBandType = GDT_Unknown;
    double       dfNoData = 0;
    int          bSuccess = 0;

    poSrcBand = poSrcDS->GetRasterBand( nBand );

    /* -------------------------------------------------------------------- */
    /*      Select output data type to match source.                        */
    /* -------------------------------------------------------------------- */
    eBandType = poSrcBand->GetRasterDataType();

    /* -------------------------------------------------------------------- */
    /*      Create this band.                                               */
    /* -------------------------------------------------------------------- */
    if( CE_Failure == poDstDS->AddBand( eBandType, NULL ) )
        return CE_Failure;

    poVRTBand = static_cast<VRTSourcedRasterBand*>( poDstDS->GetRasterBand( nBand ) );

    /* -------------------------------------------------------------------- */
    /*      Create a simple data source.                                    */
    /* -------------------------------------------------------------------- */
    poVRTBand->AddSimpleSource( poSrcBand, nSrcXOff, nSrcYOff,
                                nSrcXSize, nSrcYSize, 0, 0,
                                nSrcXSize, nSrcYSize );

    /* -------------------------------------------------------------------- */
    /*      copy over some other information of interest.                   */
    /* -------------------------------------------------------------------- */
    poVRTBand->SetMetadata( poSrcBand->GetMetadata() );
    poVRTBand->SetColorTable( poSrcBand->GetColorTable() );
    poVRTBand->SetColorInterpretation(poSrcBand->GetColorInterpretation());

    if( strlen(poSrcBand->GetDescription()) > 0 )
        poVRTBand->SetDescription( poSrcBand->GetDescription() );

    dfNoData = poSrcBand->GetNoDataValue( &bSuccess );

    if ( bSuccess )
        poVRTBand->SetNoDataValue( dfNoData );

    return CE_None;
}

/**********************************************************************/
/*                           CopyRasterWindow()                       */
/**********************************************************************/

VRTDataset* CopyRasterWindow( GDALDatasetH hSrcDS,
                                     int nSrcXOff, int nSrcYOff,
                                     int nSrcXSize, int nSrcYSize )
{
    CPLErr      err = CE_None;
    int         nBandCount = 0;
    int         i = 0;
    double      adfGeoTransform[6] = { 0 };
    const char* pszProjection = NULL;
    char**      papszMetadataOptions = NULL;
    VRTDataset *poVDS = NULL;

    // Released by the caller - SplitFile
    poVDS = new VRTDataset( nSrcXSize, nSrcYSize );

    pszProjection = GDALGetProjectionRef( hSrcDS );

    if( pszProjection != NULL && strlen(pszProjection) > 0 )
        poVDS->SetProjection( pszProjection );

    if( CE_None == GDALGetGeoTransform( hSrcDS, adfGeoTransform ) )
    {
        adfGeoTransform[0] += nSrcXOff * adfGeoTransform[1]
            + nSrcYOff * adfGeoTransform[2];
        adfGeoTransform[3] += nSrcXOff * adfGeoTransform[4]
            + nSrcYOff * adfGeoTransform[5];

        poVDS->SetGeoTransform( adfGeoTransform );
    }

    else if( GDALGetGCPCount( hSrcDS ) > 0 )
    {
        GDAL_GCP* pasGCPs = NULL;
        int nGCPs = GDALGetGCPCount( hSrcDS );

        pasGCPs = GDALDuplicateGCPs( nGCPs, GDALGetGCPs( hSrcDS ) );

        for( i = 0; i < nGCPs; i++ )
        {
            pasGCPs[i].dfGCPPixel -= nSrcXOff;
            pasGCPs[i].dfGCPLine  -= nSrcYOff;
        }

        poVDS->SetGCPs( nGCPs, pasGCPs, GDALGetGCPProjection( hSrcDS ) );

        GDALDeinitGCPs( nGCPs, pasGCPs );
        CPLFree( pasGCPs );
    }

    poVDS->SetMetadata( ((GDALDataset*)hSrcDS)->GetMetadata() );
    AttachMetadata( (GDALDatasetH) poVDS, papszMetadataOptions );

    nBandCount = GDALGetRasterCount( hSrcDS );
    for( i = 1; i <= nBandCount; i++ )
    {
        err = CopyRasterBand( (GDALDataset *) hSrcDS, poVDS, i,
                              nSrcXOff, nSrcYOff, nSrcXSize, nSrcYSize);

        CPLAssert( CE_None == err );
    }

    return poVDS;
}

/**********************************************************************/
/*                           SplitFile()                              */
/**********************************************************************/

int SplitFile( const char* pszSource, int cols, int rows,
                      int bQuiet, GDALProgressFunc pfnProgress )
{
    GDALDriverH	    hDriver = NULL;
    GDALDatasetH    hDataset = NULL;
    GDALDatasetH    hOutDS = NULL;
    VRTDataset*     poVDS = NULL;
    char*           pszDest = NULL;

    int i = 0;
    int x = 0;
    int z = 0;
    int nRasterXSize = 0;
    int nRasterYSize = 0;
    int nSrcXOff = 0;
    int nSrcYOff = 0;
    int nRoundedXSize = 0;
    int nRoundedYSize = 0;
    int nSrcXSize = 0;
    int nSrcYSize = 0;

    /* -------------------------------------------------------------------- */
    /*      Attempt to open source file.                                    */
    /* -------------------------------------------------------------------- */

    hDataset = GDALOpen( pszSource, GA_ReadOnly );

    if( NULL == hDataset )
    {
        fprintf( stderr,
            "GDALOpen failed - %d\n%s\n",
            CPLGetLastErrorNo(), CPLGetLastErrorMsg() );
        return ( 0 );
    }

    if( CSLCount(GDALGetMetadata( hDataset, "SUBDATASETS" )) > 0 )
    {
        fprintf( stderr,
            "Input file contains subdatasets. Please, select one of them for reading.\n" );
        GDALClose( hDataset );
        return ( 0 );
    }

    nRasterXSize = GDALGetRasterXSize( hDataset );
    nRasterYSize = GDALGetRasterYSize( hDataset );

    if( cols > nRasterXSize )
    {
        fprintf( stderr,
            "The column count cannot be greater than the image width in pixels.\n" );
        GDALClose( hDataset );
        return ( 0 );
    }

    if( rows > nRasterYSize )
    {
        fprintf( stderr,
            "The row count cannot be greater than the image height in pixels.\n" );
        GDALClose( hDataset );
        return ( 0 );
    }

    if( !bQuiet )
    {
        printf( "Splitting %s size %d, %d into %d files...\n",
                pszSource, nRasterXSize, nRasterYSize, cols * rows );
    }

    /* -------------------------------------------------------------------- */
    /*      Find the output driver.                                         */
    /* -------------------------------------------------------------------- */

    hDriver = GDALGetDatasetDriver( hDataset );

    /* -------------------------------------------------------------------- */
    /*      Establish some parameters.                                      */
    /* -------------------------------------------------------------------- */

    nRoundedXSize = round<int>( nRasterXSize / static_cast<double>( cols ) );
    nRoundedYSize = round<int>( nRasterYSize / static_cast<double>( rows ) );

    /* -------------------------------------------------------------------- */
    /*      Make a virtual clone.                                           */
    /* -------------------------------------------------------------------- */

    for (i = 1, z = 1, nSrcXOff = 0; i <= cols; i++, nSrcXOff += nSrcXSize)
    {
        nSrcXSize = nRoundedXSize; // (i != cols) ? nRoundedXSize : (nRasterXSize - nRoundedXSize);

        for (x = 1, nSrcYOff = 0; x <= rows; x++, z++, nSrcYOff += nSrcYSize)
        {
            pszDest = CreateFileName(pszSource, i, x);
            if (NULL == pszDest)
                return FALSE;

            nSrcYSize = nRoundedYSize; // (x != rows) ? nRoundedYSize : (nRasterYSize - nRoundedYSize);

            if (!bQuiet)
            {
                printf( "%d. %s\n offs: %d x %d\t size: %d x %d\n", z, pszDest,
                        nSrcXOff, nSrcYOff, nSrcXSize, nSrcYSize);
            }

            poVDS = CopyRasterWindow(hDataset, nSrcXOff, nSrcYOff, nSrcXSize, nSrcYSize);

            /* -------------------------------------------------------------------- */
            /*      Write to the output file using CopyCreate().                    */
            /* -------------------------------------------------------------------- */
            hOutDS = GDALCreateCopy( hDriver, pszDest, (GDALDatasetH)poVDS,
                                     TRUE, NULL, pfnProgress, NULL );
            if( NULL != hOutDS )
            {
                GDALClose( hOutDS );
            }

            /* Remove dataset allocated by CopyRasterWindow called above. */
            GDALClose( (GDALDatasetH) poVDS );

            CPLFree( pszDest );
        }
    }

    GDALClose( hDataset );

    return ( hOutDS != NULL );
}

} // anonymous namespace

/**********************************************************************/
/*                                main()                              */
/**********************************************************************/

int main( int argc, char ** argv )
{
    int bQuiet = FALSE;
    int i = 0;
    int j = 0;
    int cols = DefaultColsNumber;
    int rows = DefaultRowsNumber;
    GDALProgressFunc pfnProgress = GDALTermProgress;

    /* -------------------------------------------------------------------- */
    /*      Register standard GDAL drivers, and identify output driver.     */
    /* -------------------------------------------------------------------- */

    GDALAllRegister();

    /* -------------------------------------------------------------------- */
    /*      Handle command line arguments.                                  */
    /* -------------------------------------------------------------------- */

    for( i = 1; i < argc; i++ )
    {
        if( EQUAL(argv[i],"--version") )
        {
            printf( "%s\n", GDALVersionInfo( "--version" ) );
        }

        else if( EQUAL(argv[i],"-quiet") )
        {
            bQuiet = TRUE;
            pfnProgress = GDALDummyProgress;
        }	

        else if( EQUAL(argv[i],"-c") )
        {
            if( atoi(argv[i+1]) < 2 )
            {
                printf( "There must be more columns than %s.\n", argv[i+1] );
                Usage();
                GDALDestroyDriverManager();
                exit( 2 );
            }

            cols = atoi(argv[++i]);
        }

        else if( EQUAL(argv[i],"-r") )
        {
            if( atoi(argv[i+1]) < 2 )
            {
                printf( "There must be more rows than %s.\n", argv[i+1] );
                Usage();
                GDALDestroyDriverManager();
                exit( 2 );
            }

            rows = atoi(argv[++i]);
        }

        else if( argv[i][0] == '-' )
        {
            printf( "Option %s incomplete, or not recognised.\n\n", 
                argv[i] );
            Usage();
            GDALDestroyDriverManager();
            exit( 2 );
        }

        else
        {
            SplitFile( argv[i], cols, rows, bQuiet, pfnProgress );
            j++;
        }
    }

    /* At least one file must be provided */
    if (j == 0)
    {
        Usage();
        GDALDestroyDriverManager();
        exit( 2 );
    }

    /* -------------------------------------------------------------------- */
    /*      Clean up.                                                       */
    /* -------------------------------------------------------------------- */

    GDALDumpOpenDatasets( stderr );
    GDALDestroyDriverManager();

    exit(0);
}

